% -*- SLang -*-         ruby.sl

% Ruby mode for Jed
% by MAEDA Shugo (ender@pic-internet.or.jp)

% `Ruby mode for Jed' is FREESOFTWARE.
% Please use AT YOUR OWN RISK.

% What's ruby?
%
%  Ruby is the interpreted scripting language for quick and
%  easy object-oriented programming.  It has many features to
%  process text files and to do system management tasks (as in
%  perl).  It is simple, straight-forward, and extensible.
%
%  The ruby distribution can be found on
% 
%    ftp://ftp.caelum.co.jp/pub/lang/ruby/

% Install
%
% Please add these lines to your `.jedrc' file.
%
%   ~/.jedrc
%
%     % amount of space to indent within block.
%     variable ruby_indent_level = 2;
%
%     % Load ruby mode when openning `.rb' files.   
%     autoload("ruby_mode", "ruby");
%     add_mode_for_extension ("ruby", "rb");

variable ruby_mode_version = "0.0.2";

!if (is_defined ("ruby_indent_level"))
{
   variable ruby_indent_level = 2;
}

define ruby_show_version()
{
   message(Sprintf("ruby mode for Jed version %s", ruby_mode_version, 1));
}

define ruby_indent_to(n)
{
   variable step;
   
   step = what_column();
   bol_skip_white();
   step -= what_column();
   if (what_column != n)
     {
	bol_trim ();
	n--;
	whitespace (n);
     }
   if (step > 0) go_right(step);
}

define ruby_looking_keyword_at(keyword)
{
   push_spot;
   EXIT_BLOCK
     {
	pop_spot;
     }
   
   if (looking_at(keyword))
     {
	go_right(strlen(keyword));
	return (orelse
		 { looking_at(" ") }
		 { looking_at("\t") }
		 { looking_at(";") }
		 { eolp() }
		)
     }
   else
     {
	return 0;
     }
}
	  
define ruby_calculate_indent()
{
   variable indent = 0;
   variable extra_indent = 0;
   variable ch;
   variable par_level;
   
   CASE_SEARCH = 0;
   push_spot();
   EXIT_BLOCK
     {
	pop_spot();
     }
   
   bol_skip_white();
   indent = what_column();
   if (orelse
	{ ruby_looking_keyword_at("end") }
	{ ruby_looking_keyword_at("else") }
	{ ruby_looking_keyword_at("elsif") }
	{ ruby_looking_keyword_at("rescue") }
	{ ruby_looking_keyword_at("ensure") }
	{ ruby_looking_keyword_at("when") }
	{ looking_at("}") }
       )
     {
	extra_indent -= ruby_indent_level;
     }
   !if (up_1()) return indent;

   eol();
   par_level = 0;
   forever
     {
	if (eolp())
	  {
	     forever
	       {
		  bol();
		  if (looking_at("#"))
		    {
		       !if (up_1()) return indent;
		       eol();
		    }
		  else
		    {
		       eol();
		       break;
		    }
	       }
	  }
	go_left_1();
	ch = what_char();
	if (ch == ')')
	  {
	     par_level--;
	  }
	else if (ch == '(')
	  {
	     par_level++;
	     if (par_level == 1) return what_column() + 1;
	  }
	
	if (bolp() and (par_level == 0))
	  {
	     skip_white();
	     indent = what_column();
	     break;
	  }
     }
   
   if (looking_at("#")) return what_column();
   
   if (orelse
	{ ruby_looking_keyword_at("class") }
	{ ruby_looking_keyword_at("module") }
	{ ruby_looking_keyword_at("def") }
	{ ruby_looking_keyword_at("if") }
	{ ruby_looking_keyword_at("else") }
	{ ruby_looking_keyword_at("elsif") }
	{ ruby_looking_keyword_at("unless") }
	{ ruby_looking_keyword_at("case") }
	{ ruby_looking_keyword_at("when") }
	{ ruby_looking_keyword_at("while") }
	{ ruby_looking_keyword_at("until") }
	{ ruby_looking_keyword_at("for") }
	{ ruby_looking_keyword_at("begin") }
	{ ruby_looking_keyword_at("rescue") }
	{ ruby_looking_keyword_at("ensure") }
       )
     {
	eol();
	bskip_white();
	!if (orelse
	      { blooking_at(" end") }
	      { blooking_at("\tend") }
	     ) 
	  {
	     extra_indent += ruby_indent_level;
	  }
     }
   else
     {
	eol();
	bskip_white();
	if (blooking_at("{"))
	  extra_indent += ruby_indent_level;
	else if (blooking_at("|"))
	  extra_indent += ruby_indent_level;
	else if (blooking_at(" do"))
	  extra_indent += ruby_indent_level;
     }
   
   return indent + extra_indent;
}

define ruby_indent_line()
{
   ruby_indent_to(ruby_calculate_indent());
}

define ruby_newline_and_indent()
{
   variable step;
   step = what_column();
   bol_skip_white();
   step -= what_column();
   if (orelse
	{ looking_at("end") }
	{ looking_at("else") }
	{ looking_at("elsif") }
	{ looking_at("rescue") }
	{ looking_at("ensure") }
	{ looking_at("when") }
	{ looking_at("}") }
       )
     {
	ruby_indent_line();
     }
   go_right(step);
   newline();
   ruby_indent_line();
}

define ruby_insert(s)
{
   variable step;
   
   insert(s);
   step = what_column();
   bol_skip_white();
   step -= what_column();
   if (orelse
	{ looking_at("end") }
	{ looking_at("else") }
	{ looking_at("elsif") }
	{ looking_at("rescue") }
	{ looking_at("ensure") }
	{ looking_at("when") }
	{ looking_at("}") }
       )
     {
	ruby_indent_line();
     }
   go_right(step);
}

% Define keymap.
$1 = "ruby";
!if (keymap_p($1)) make_keymap($1);
definekey ("ruby_show_version", "^Cv", $1);
definekey ("ruby_insert(\"0\")", "0", $1);
definekey ("ruby_insert(\"1\")", "1", $1);
definekey ("ruby_insert(\"2\")", "2", $1);
definekey ("ruby_insert(\"3\")", "3", $1);
definekey ("ruby_insert(\"4\")", "4", $1);
definekey ("ruby_insert(\"5\")", "5", $1);
definekey ("ruby_insert(\"6\")", "6", $1);
definekey ("ruby_insert(\"7\")", "7", $1);
definekey ("ruby_insert(\"8\")", "8", $1);
definekey ("ruby_insert(\"9\")", "9", $1);
definekey ("ruby_insert(\"a\")", "a", $1);
definekey ("ruby_insert(\"b\")", "b", $1);
definekey ("ruby_insert(\"c\")", "c", $1);
definekey ("ruby_insert(\"d\")", "d", $1);
definekey ("ruby_insert(\"e\")", "e", $1);
definekey ("ruby_insert(\"f\")", "f", $1);
definekey ("ruby_insert(\"g\")", "g", $1);
definekey ("ruby_insert(\"h\")", "h", $1);
definekey ("ruby_insert(\"i\")", "i", $1);
definekey ("ruby_insert(\"j\")", "j", $1);
definekey ("ruby_insert(\"k\")", "k", $1);
definekey ("ruby_insert(\"l\")", "l", $1);
definekey ("ruby_insert(\"m\")", "m", $1);
definekey ("ruby_insert(\"n\")", "n", $1);
definekey ("ruby_insert(\"o\")", "o", $1);
definekey ("ruby_insert(\"p\")", "p", $1);
definekey ("ruby_insert(\"q\")", "q", $1);
definekey ("ruby_insert(\"r\")", "r", $1);
definekey ("ruby_insert(\"s\")", "s", $1);
definekey ("ruby_insert(\"t\")", "t", $1);
definekey ("ruby_insert(\"u\")", "u", $1);
definekey ("ruby_insert(\"v\")", "v", $1);
definekey ("ruby_insert(\"w\")", "w", $1);
definekey ("ruby_insert(\"x\")", "x", $1);
definekey ("ruby_insert(\"y\")", "y", $1);
definekey ("ruby_insert(\"z\")", "z", $1);
definekey ("ruby_insert(\"A\")", "A", $1);
definekey ("ruby_insert(\"B\")", "B", $1);
definekey ("ruby_insert(\"C\")", "C", $1);
definekey ("ruby_insert(\"D\")", "D", $1);
definekey ("ruby_insert(\"E\")", "E", $1);
definekey ("ruby_insert(\"F\")", "F", $1);
definekey ("ruby_insert(\"G\")", "G", $1);
definekey ("ruby_insert(\"H\")", "H", $1);
definekey ("ruby_insert(\"I\")", "I", $1);
definekey ("ruby_insert(\"J\")", "J", $1);
definekey ("ruby_insert(\"K\")", "K", $1);
definekey ("ruby_insert(\"L\")", "L", $1);
definekey ("ruby_insert(\"M\")", "M", $1);
definekey ("ruby_insert(\"N\")", "N", $1);
definekey ("ruby_insert(\"O\")", "O", $1);
definekey ("ruby_insert(\"P\")", "P", $1);
definekey ("ruby_insert(\"Q\")", "Q", $1);
definekey ("ruby_insert(\"R\")", "R", $1);
definekey ("ruby_insert(\"S\")", "S", $1);
definekey ("ruby_insert(\"T\")", "T", $1);
definekey ("ruby_insert(\"U\")", "U", $1);
definekey ("ruby_insert(\"V\")", "V", $1);
definekey ("ruby_insert(\"W\")", "W", $1);
definekey ("ruby_insert(\"X\")", "X", $1);
definekey ("ruby_insert(\"Y\")", "Y", $1);
definekey ("ruby_insert(\"Z\")", "Z", $1);
definekey ("ruby_insert(\"_\")", "_", $1);
definekey ("ruby_insert(\"{\")", "{", $1);
definekey ("ruby_insert(\"}\")", "}", $1);
definekey ("ruby_insert(\";\")", ";", $1);

% Create syntax table.
create_syntax_table ($1);
define_syntax ("#", Null_String, '%', $1);
define_syntax ("([{<", ")]}>", '(', $1);
define_syntax ('"', '"', $1);
define_syntax ('\'', '\'', $1);
define_syntax ('\\', '\\', $1);
define_syntax ("$0-9a-zA-Z_", 'w', $1);
define_syntax ("-+0-9a-fA-F.xXL", '0', $1);
define_syntax (",;.?:", ',', $1);
define_syntax ("%-+/&*=<>|!~^", '+', $1);
set_syntax_flags ($1, 4);

#ifdef HAS_DFA_SYNTAX
enable_highlight_cache("ruby.dfa", $1);
define_highlight_rule("#.*$", "comment", $1);
define_highlight_rule("([\\$%&@\\*]|\\$#)[A-Za-z_0-9]+", "normal", $1);
define_highlight_rule(strcat("\\$([_\\./,\"\\\\#\\*\\?\\]\\[;!@:\\$<>\\(\\)",
			     "%=\\-~\\^\\|&`'\\+]|\\^[A-Z])"), "normal", $1);
define_highlight_rule("[A-Za-z_][A-Za-z_0-9]*", "Knormal", $1);
define_highlight_rule("[0-9]+(\\.[0-9]+)?([Ee][\\+\\-]?[0-9]*)?", "number",
		      $1);
define_highlight_rule("0[xX][0-9A-Fa-f]*", "number", $1);
define_highlight_rule("[\\(\\[\\{\\<\\>\\}\\]\\),;\\.\\?:]", "delimiter", $1);
define_highlight_rule("[%\\-\\+/&\\*=<>\\|!~\\^]", "operator", $1);
define_highlight_rule("-[A-Za-z]", "keyword0", $1);
define_highlight_rule("'[^']*'", "string", $1);
define_highlight_rule("'[^']*$", "string", $1);
define_highlight_rule("\"([^\"\\\\]|\\\\.)*\"", "string", $1);
define_highlight_rule("\"([^\"\\\\]|\\\\.)*\\\\?$", "string", $1);
define_highlight_rule("m?/([^/\\\\]|\\\\.)*/[gio]*", "string", $1);
define_highlight_rule("m/([^/\\\\]|\\\\.)*\\\\?$", "string", $1);
define_highlight_rule("s/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?/[geio]*",
		      "string", $1);
define_highlight_rule("s/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?\\\\?$",
		      "string", $1);
define_highlight_rule("(tr|y)/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?/[cds]*",
		      "string", $1);
define_highlight_rule("(tr|y)/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?\\\\?$",
		      "string", $1);
define_highlight_rule(".", "normal", $1);
build_highlight_table ($1);
#endif

% Type 0 keywords
() = define_keywords_n ($1, "doifinor", 2, 0);
() = define_keywords_n ($1, "anddefendfornilnot", 3, 0);
() = define_keywords_n ($1, "caseelsefailloadloopnextredoselfthenwhen", 4, 0);
() = define_keywords_n ($1, "aliasbeginbreakclasselsifraiseretrysuperundefuntilwhileyield", 5, 0);
() = define_keywords_n ($1, "ensuremodulerescuereturnunless", 6, 0);
() = define_keywords_n ($1, "includerequire", 7, 0);
() = define_keywords_n ($1, "autoload", 8, 0);
% Type 1 keywords (commonly used libc functions)
() = define_keywords_n($1, "TRUE", 4, 1);
() = define_keywords_n($1, "FALSE", 5, 1);

define ruby_mode()
{
   variable kmap = $1;
   set_mode(kmap, 2);
   use_keymap(kmap);
   use_syntax_table(kmap);
   set_buffer_hook("indent_hook", "ruby_indent_line");
   set_buffer_hook("newline_indent_hook", "ruby_newline_and_indent"); 
   runhooks("ruby_mode_hook");
}
